import os
import re
import sys
from functools import cache
from typing import TYPE_CHECKING, Self

from kleur import Color

if TYPE_CHECKING:
    from collections.abc import Callable

    from kleur import RGB

ANSI_ESCAPE = "\x1b"
_ANSI_STYLE_REGEX = re.compile(rf"{ANSI_ESCAPE}\[\d+(;\d+)*m")


def ansi(s: str) -> str:
    return f"{ANSI_ESCAPE}[{s}"


def ansi_style(*values: int) -> str:
    return ansi(f"{';'.join(str(v) for v in values)}m")


LINE_UP = ansi("A")
LINE_CLEAR = ansi("2K")

RESET_STYLE = ansi_style(0)


@cache
def has_colors() -> bool:
    no = "NO_COLOR" in os.environ
    yes = "CLICOLOR_FORCE" in os.environ
    maybe = sys.stdout.isatty()
    return not no and (yes or maybe)


def _apply_ansi_style(s: str, *values: int) -> str:
    if values and has_colors():
        return f"{ansi_style(*values)}{s}{RESET_STYLE}"
    return s


type _StringStyler = Callable[[str], str]


def _wrap_ansi_style(*values: int) -> _StringStyler:
    def wrapper(s: str) -> str:
        return _apply_ansi_style(s, *values)

    return wrapper


def strip_ansi_style(s: str) -> str:
    return _ANSI_STYLE_REGEX.sub("", s)


bold = _wrap_ansi_style(1)
faint = _wrap_ansi_style(2)
italic = _wrap_ansi_style(3)
underlined = _wrap_ansi_style(4)
inverse = _wrap_ansi_style(7)
strikethrough = _wrap_ansi_style(9)

black = _wrap_ansi_style(30)
red = _wrap_ansi_style(31)
green = _wrap_ansi_style(32)
yellow = _wrap_ansi_style(33)
blue = _wrap_ansi_style(34)
magenta = _wrap_ansi_style(35)
cyan = _wrap_ansi_style(36)
gray = _wrap_ansi_style(37)

black_background = _wrap_ansi_style(40)
red_background = _wrap_ansi_style(41)
green_background = _wrap_ansi_style(42)
yellow_background = _wrap_ansi_style(43)
blue_background = _wrap_ansi_style(44)
magenta_background = _wrap_ansi_style(45)
cyan_background = _wrap_ansi_style(46)
gray_background = _wrap_ansi_style(47)

light_gray = _wrap_ansi_style(90)
light_red = _wrap_ansi_style(91)
light_green = _wrap_ansi_style(92)
light_yellow = _wrap_ansi_style(93)
light_blue = _wrap_ansi_style(94)
light_magenta = _wrap_ansi_style(95)
light_cyan = _wrap_ansi_style(96)
white = _wrap_ansi_style(97)

OK = green("✔")
FAIL = red("✘")


def color_8bit(fg: int = None, bg: int = None) -> _StringStyler:
    values = []
    if fg:
        values += [38, 5, fg]
    if bg:
        values += [48, 5, bg]
    return _wrap_ansi_style(*values)


def color_rgb(fg: RGB = None, bg: RGB = None) -> _StringStyler:
    values = []
    if fg:
        values += [38, 2, *fg]
    if bg:
        values += [48, 2, *bg]
    return _wrap_ansi_style(*values)


class Colored[T](str):
    value: T
    bg: Color | None
    fg: Color | None
    __slots__ = "bg", "fg", "value"

    def __new__(cls, value: T, fg: Color = None, bg: Color = None) -> Self:
        fg_rgb, bg_rgb = fg.as_rgb if fg else None, bg.as_rgb if bg else None
        instance = super().__new__(cls, color_rgb(fg_rgb, bg_rgb)(str(value)))
        instance.value, instance.fg, instance.bg = (value, fg, bg)
        return instance

    def with_color(self, color: Color) -> Colored:
        return Colored(self.value, color, self.bg)

    def with_background(self, background: Color) -> Colored:
        return Colored(self.value, self.fg, background)


class Highlighter:
    def __init__(self, color: Color) -> None:
        self._color = color

    def __call__(self, s: str, *, inverted: bool = False, enabled: bool = True) -> str:
        if not enabled:
            return s
        c, k = self._color.contrasting_shade_pair
        return Colored(s, *((c, k) if inverted else (k, c)))


CP = Color.Props


class ColorHighlighter:
    def __init__(self, color: Color) -> None:
        self._color = color

    def __call__(
        self, highlighted: CP = CP.ALL, *, enable_bounds_highlights: bool = False
    ) -> str:
        c, highlighter = self._color, Highlighter(self._color)
        # Colors progressively built up with hue, saturation & lightness
        decomposed = [c.with_props(CP.H), c.with_props(CP.NO_L), c]
        # Go over each color property and highlight it if necessary.
        sh, ss, sl = [
            Highlighter(k)(f" {s} ", enabled=p in highlighted)
            for (s, k, p) in zip(c.prop_strings(), decomposed, iter(CP), strict=True)
        ]
        # Highlight the outer brackets (or don't), to make it more clearly
        # noticeable if a color is highlighted.
        is_hl = bool(highlighted) and enable_bounds_highlights
        hsluv, start, end = [highlighter(s, enabled=is_hl) for s in ("HSLuv", "[", "]")]
        return f"{hsluv} {start} {sh} {ss} {sl} {end}"



from dataclasses import dataclass, replace
from enum import IntFlag, auto
from functools import cached_property, total_ordering
from typing import TYPE_CHECKING, NamedTuple

from hsluv import hex_to_hsluv, hsluv_to_hex, hsluv_to_rgb, rgb_to_hsluv

from .interpol import mapped, mapped_cyclic, trim, trim_cyclic

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

_INCREASE_FACTOR = 2**0.5


class HasNormalizeArgs:
    def _normalize_values(self, funcs: dict[str, Callable[[float], float]]) -> None:
        """
        Make sure all color values are in a valid range.

        object.__setattr__() is one of the awkward options (*) we have,
        when we want to set attributes in a frozen dataclass (which will raise a
        FrozenInstanceError when its own __setattr__() or __delattr__() is invoked).

        *) Another option could be to move the attributes to a super class and
        call super().__init__() here.
        """
        for name, func in funcs.items():
            object.__setattr__(self, name, func(getattr(self, name)))


def normalize_rgb_hex(rgb_hex: str) -> str:
    """
    Try to normalize a hex string into a rrggbb hex.

    :param rgb_hex: RGB hex string (may start with '#')
    :return: rrggbb hex

    >>> normalize_rgb_hex("3")
    '333333'
    >>> normalize_rgb_hex("03")
    '030303'
    >>> normalize_rgb_hex("303")
    '330033'
    >>> normalize_rgb_hex("808303")
    '808303'
    """
    rgb_hex, r, g, b = rgb_hex.removeprefix("#").lower(), "", "", ""

    match len(rgb_hex):
        case 1:
            # 3 -> r=33, g=33, b=33
            r = g = b = rgb_hex * 2

        case 2:
            # 03 -> r=03, g=03, b=03
            r = g = b = rgb_hex

        case 3:
            # 303 -> r=33, g=00, b=33
            r1, g1, b1 = iter(rgb_hex)
            r, g, b = r1 * 2, g1 * 2, b1 * 2

        case 6:
            # 808303 -> r=80, g=83, b=03
            r1, r2, g1, g2, b1, b2 = iter(rgb_hex)
            r, g, b = r1 + r2, g1 + g2, b1 + b2

        case _:
            raise ValueError(rgb_hex)

    return f"{r}{g}{b}"


type RGB = tuple[int, int, int]


class _HSLuv(NamedTuple):
    hue: float
    saturation: float
    lightness: float

    @classmethod
    def from_hex(cls, rgb_hex: str) -> _HSLuv:
        return cls(*hex_to_hsluv(f"#{rgb_hex}"))

    @property
    def as_hex(self) -> str:
        return hsluv_to_hex(self)[1:]

    @classmethod
    def from_rgb(cls, rgb: RGB) -> _HSLuv:
        r, g, b = rgb
        return cls(*rgb_to_hsluv((r / 255, g / 255, b / 255)))

    @property
    def as_rgb(self) -> RGB:
        r, g, b = hsluv_to_rgb(self)
        return round(r * 255), round(g * 255), round(b * 255)


@total_ordering
@dataclass(frozen=True)
class Color(HasNormalizeArgs):
    class Props(IntFlag):
        H = auto()
        S = auto()
        L = auto()
        NONE = 0
        NO_L = H | S
        NO_S = H | L
        ALL = H | S | L

    hue: float = 0  # 0 - 1 (full circle angle)
    saturation: float = 1  # 0 - 1 (ratio)
    lightness: float = 0.5  # 0 - 1 (ratio)

    def __post_init__(self) -> None:
        self._normalize_values(
            {"hue": trim_cyclic, "saturation": trim, "lightness": trim}
        )

    def __repr__(self) -> str:
        sh, ss, sl = self.prop_strings()
        return f"HSLuv({sh}, {ss}, {sl})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Color):
            return self.as_rgb == other.as_rgb
        raise NotImplementedError

    def __hash__(self) -> int:
        return hash(iter(self))

    def __lt__(self, other: Color) -> bool:
        return self.as_sortable_tuple < other.as_sortable_tuple

    def __iter__(self) -> Iterator[float]:
        yield self.hue
        yield self.saturation
        yield self.lightness

    def with_props(self, props: Props) -> Color:
        """
        Color built up with from a selection of its original hue, saturation, lightness.

        This could be helpful for understanding how colors
        are built up and relate to each other.
        """
        prop_values = zip(iter(self), iter(self.Props), strict=True)
        return Color(*[v for v, p in prop_values if p in props])

    def prop_strings(self) -> Iterator[str]:
        for v, s in zip(self._as_hsluv, ("°", "%", "%"), strict=True):
            yield f"{v:.2f}{s}".rjust(7)

    def adjust(
        self, *, hue: float = None, saturation: float = None, lightness: float = None
    ) -> Color:
        return Color(
            self.hue + (hue or 0),
            self.saturation * (saturation or 1),
            self.lightness * (lightness or 1),
        )

    def align(self, other: Color) -> Color:
        is_unsaturated = round(self.saturation, 2) == 0
        return self.shifted(other.hue) if is_unsaturated else self

    def align_pair(self, other: Color) -> tuple[Color, Color]:
        aligned_self = self.align(other)
        return aligned_self, other.align(aligned_self)

    @cached_property
    def as_sortable_tuple(self) -> tuple[float, float, float]:
        """Will decide the sort order."""
        return self.lightness, self.saturation, self.hue

    @classmethod
    def _from_hsluv(cls, hsluv: _HSLuv) -> Color:
        return cls(hsluv.hue / 360, hsluv.saturation / 100, hsluv.lightness / 100)

    @cached_property
    def _as_hsluv(self) -> _HSLuv:
        return _HSLuv(self.hue * 360, self.saturation * 100, self.lightness * 100)

    @classmethod
    def from_hex(cls, rgb_hex: str) -> Color:
        """
        Create a Color from an RGB hex string.

        :param rgb_hex: RGB hex string (may start with '#')
        :return: Color instance

        >>> c = Color.from_hex("808303")
        >>> c.as_hex, c.as_rgb
        ('808303', (128, 131, 3))
        >>> k = Color.from_hex("0af")
        >>> k.as_hex, k.as_rgb
        ('00aaff', (0, 170, 255))
        """
        return cls._from_hsluv(_HSLuv.from_hex(normalize_rgb_hex(rgb_hex)))

    @cached_property
    def as_hex(self) -> str:
        return self._as_hsluv.as_hex

    @classmethod
    def from_rgb(cls, rgb: RGB) -> Color:
        """
        Create a Color from RGB values.

        :param rgb: RGB instance
        :return: Color instance

        >>> c = Color.from_rgb((128, 131, 3))
        >>> c.as_hex, c.as_rgb
        ('808303', (128, 131, 3))
        >>> k = Color.from_rgb((0, 170, 255))
        >>> k.as_hex, k.as_rgb
        ('00aaff', (0, 170, 255))
        """
        return cls._from_hsluv(_HSLuv.from_rgb(rgb))

    @cached_property
    def as_rgb(self) -> RGB:
        return self._as_hsluv.as_rgb

    def shifted(self, hue: float) -> Color:
        return replace(self, hue=hue)

    def saturated(self, saturation: float) -> Color:
        return replace(self, saturation=saturation)

    def shade(self, lightness: float) -> Color:
        return replace(self, lightness=lightness)

    def shades(self, n_intervals: int) -> Iterator[Color]:
        for step in range(1, n_intervals):
            yield self.shade(step / n_intervals)

    @cached_property
    def very_bright(self) -> Color:
        return self.shade(5 / 6)

    @cached_property
    def bright(self) -> Color:
        return self.shade(4 / 6)

    @cached_property
    def dark(self) -> Color:
        return self.shade(2 / 6)

    @cached_property
    def very_dark(self) -> Color:
        return self.shade(1 / 6)

    def brighter(self, relative_amount: float = _INCREASE_FACTOR) -> Color:
        return self.adjust(lightness=relative_amount)

    @cached_property
    def slightly_brighter(self) -> Color:
        return self.brighter(_INCREASE_FACTOR**0.5)

    @cached_property
    def much_brighter(self) -> Color:
        return self.brighter(_INCREASE_FACTOR**2)

    def darker(self, relative_amount: float = _INCREASE_FACTOR) -> Color:
        return self.brighter(1 / relative_amount)

    @cached_property
    def slightly_darker(self) -> Color:
        return self.darker(_INCREASE_FACTOR**0.5)

    @cached_property
    def much_darker(self) -> Color:
        return self.darker(_INCREASE_FACTOR**2)

    def blend(self, other: Color, amount: float = 0.5) -> Color:
        return Color(
            mapped_cyclic(amount, (self.hue, other.hue), period=1),
            mapped(amount, (self.saturation, other.saturation)),
            mapped(amount, (self.lightness, other.lightness)),
        )

    @cached_property
    def contrasting_shade(self) -> Color:
        """
        Color with a lightness that contrasts with the current color.

        Color with a 50% lower or higher lightness than the current color,
        while maintaining the same hue and saturation (so it can for example
        be used as background color).

        :return: Color representation of the contrasting shade

        >>> hex_strs = ["08f", "0f8", "80f", "8f0", "f08", "f80"]
        >>> for c, k in [Color.from_hex(h).contrasting_shade_pair for h in hex_strs]:
        ...     print(f"{c.as_hex} <-> {k.as_hex}")
        0088ff <-> 001531
        00ff88 <-> 006935
        8800ff <-> ebe4ff
        88ff00 <-> 366b00
        ff0088 <-> 2b0012
        ff8800 <-> 4a2300
        """
        return self.shade((self.lightness + 0.5) % 1)

    @cached_property
    def contrasting_shade_pair(self) -> tuple[Color, Color]:
        """
        Return this color together with its contrasting shade.

        Turns out to be useful quite commonly in practice, especially in situations
        when you'd want to accommodate a color with a background that would deliver the
        best contrast in terms of visibility.
        """
        return self, self.contrasting_shade

    @cached_property
    def contrasting_hue(self) -> Color:
        """
        Color with a hue that contrasts with the current color.

        Color with a 180° different hue than the current color,
        while maintaining the same saturation and perceived lightness.

        :return: Color representation of the contrasting hue

        >>> hex_strs = ["08f", "0f8", "80f", "8f0", "f08", "f80"]
        >>> for c, k in [Color.from_hex(h).contrasting_hue_pair for h in hex_strs]:
        ...     print(f"{c.as_hex} <-> {k.as_hex}")
        0088ff <-> 9c8900
        00ff88 <-> ffd1f5
        8800ff <-> 5c6900
        88ff00 <-> f6d9ff
        ff0088 <-> 009583
        ff8800 <-> 00b8d1
        """
        return self.adjust(hue=0.5)

    @cached_property
    def contrasting_hue_pair(self) -> tuple[Color, Color]:
        """Return this color together with its contrasting hue."""
        return self, self.contrasting_hue


def blend_colors(c: Color, k: Color) -> Callable[[float], Color]:
    def wrapped(amount: float) -> Color:
        return c.blend(k, amount)

    return wrapped



from functools import cached_property
from math import log, pi

FULL_CIRCLE = pi * 2


type _Bounds = tuple[float, float]


def trim(n: float, lower: float = 0, upper: float = 1) -> float:
    return min(max(lower, n), upper)


def trim_cyclic(n: float, period: float = 1) -> float:
    return n % period


class _MappingBounds:
    def __init__(self, start: float = 0, end: float = 1) -> None:
        self._start = start
        self._end = end

    @cached_property
    def _span(self) -> float:
        return self._end - self._start

    def value_at(self, f: float) -> float:
        return self._start + self._span * f

    def position_of(self, n: float) -> float:
        try:
            return (n - self._start) / self._span
        except ZeroDivisionError:
            return 0.0


class LinearMapping(_MappingBounds):
    def position_of(self, n: float, *, inside: bool = False) -> float:
        f = super().position_of(n)
        return trim(f, 0.0, 1.0) if inside else f


class LogarithmicMapping(LinearMapping):
    def __init__(self, start: float = 0, end: float = 1, base: float = 10) -> None:
        super().__init__(log(start, base), log(end, base))
        self._base = base

    def value_at(self, f: float) -> float:
        return self._base ** super().value_at(f)

    def position_of(self, n: float, *, inside: bool = False) -> float:
        return super().position_of(log(n, self._base), inside=inside)


class CyclicMapping(_MappingBounds):
    def __init__(
        self, start: float = 0, end: float = FULL_CIRCLE, period: float = FULL_CIRCLE
    ) -> None:
        self._period = period
        start, end = self._trim(start), self._trim(end)

        # To ensure interpolation over the smallest angle,
        # phase shift {start} over whole periods, such that the
        # (absolute) difference between {start} <-> {end} <= 1/2 {period}.
        #
        #                          v------ period ------v
        #    -1                    0                    1                    2
        #     |                    |                    |     start < end:   |
        # Old:|                    |   B ~~~~~~~~~> E   |                    |
        # New:|                    |                E <~|~~ B' = B + period  |
        #     |    start > end:    |                    |                    |
        # Old:|                    |   E <~~~~~~~~~ B   |                    |
        # New:|  B - period =  B'~~|~> E                |                    |

        if abs(end - start) > period / 2:
            start += period if start < end else -period

        super().__init__(start, end)

    def _trim(self, n: float) -> float:
        return n % self._period

    def value_at(self, f: float) -> float:
        return self._trim(super().value_at(f))

    def position_of(self, n: float) -> float:
        return super().position_of(self._trim(n))


def mapped(f: float, bounds: _Bounds) -> float:
    return LinearMapping(*bounds).value_at(f)


def unmapped(n: float, bounds: _Bounds, *, inside: bool = False) -> float:
    return LinearMapping(*bounds).position_of(n, inside=inside)


def mapped_log(f: float, bounds: _Bounds, *, base: float = 10) -> float:
    return LogarithmicMapping(*bounds, base).value_at(f)


def unmapped_log(
    n: float, bounds: _Bounds, *, base: float = 10, inside: bool = False
) -> float:
    return LogarithmicMapping(*bounds, base).position_of(n, inside=inside)


def mapped_cyclic(f: float, bounds: _Bounds, *, period: float = FULL_CIRCLE) -> float:
    return CyclicMapping(*bounds, period).value_at(f)


def mapped_angle(f: float, bounds: _Bounds) -> float:
    """Shorthand for mapped_cyclic() for an angle (in radians) as period."""
    return mapped_cyclic(f, bounds, period=FULL_CIRCLE)


def unmapped_cyclic(n: float, bounds: _Bounds, *, period: float = FULL_CIRCLE) -> float:
    return CyclicMapping(*bounds, period).position_of(n)


def unmapped_angle(n: float, bounds: _Bounds) -> float:
    """Shorthand for unmapped_cyclic() for an angle (in radians) as period."""
    return unmapped_cyclic(n, bounds, period=FULL_CIRCLE)


class NumberMapping:
    def __init__(self, from_bounds: _MappingBounds, to_bounds: _MappingBounds) -> None:
        self._from_bounds = from_bounds
        self._to_bounds = to_bounds

    def map(self, n: float) -> float:
        return self._to_bounds.value_at(self._from_bounds.position_of(n))


def map_number(
    n: float, from_bounds: _MappingBounds, to_bounds: _MappingBounds
) -> float:
    return NumberMapping(from_bounds, to_bounds).map(n)



from .color import RGB, Color, blend_colors, normalize_rgb_hex
from .formatting import Colored, ColorHighlighter, Highlighter
from .themes import BLACK, GREY, WHITE, AltColors, Colors, c

__all__ = [
    "BLACK",
    "GREY",
    "RGB",
    "WHITE",
    "AltColors",
    "Color",
    "ColorHighlighter",
    "Colored",
    "Colors",
    "Highlighter",
    "blend_colors",
    "c",
    "normalize_rgb_hex",
]



from kleur import Color


def c(h: int) -> Color:
    return Color(h / 360)


GREY = Color(saturation=0)
BLACK = GREY.shade(0)
WHITE = GREY.shade(1)


class Colors:
    """
    Highly opinionated (though carefully selected) color theme.

    Custom color themes can by specified by creating a very basic class
    that solely consists of a different set of hues (in degrees), for example:
    >>> from math import pi
    >>> class MyColors:
    ...     tomato = c(15)
    ...     turquoise = c(175)
    ...     very_random = tomato.blend(turquoise, 1 / pi)
    >>> for name in ["tomato", "turquoise", "very_random"]:
    ...     print(f"{getattr(MyColors, name)} <-- {name}")
    HSLuv( 15.00°, 100.00%,  50.00%) <-- tomato
    HSLuv(175.00°, 100.00%,  50.00%) <-- turquoise
    HSLuv( 65.93°, 100.00%,  50.00%) <-- very_random
    """

    red = c(12)
    orange = c(33)
    yellow = c(69)
    poison = c(101)
    green = c(127)
    ocean = c(190)
    blue = c(248)
    indigo = c(267)
    purple = c(281)
    pink = c(329)

    brown = orange.blend(yellow, 0.25).saturated(0.69)


class AltColors:
    """Alternative color theme."""

    red = c(10)
    orange = c(35)
    yellow = c(75)
    poison = c(100)
    green = c(126)
    ocean = c(184)
    blue = c(242)
    indigo = c(268)
    purple = c(280)
    pink = c(325)

    brown = orange.blend(yellow, 0.3).saturated(0.69)
